#include <iostream>
#include <cmath>
#include <vector>
#include <algorithm>
#include <cstdlib> //For rand
using namespace std;

/*
We want to invert the function y = x*exp(x). That is, we want to input y
and find the corresponding x. This program calculates y for x between 0
and 100 in steps of 1e-5 and saves the values in an array (usually
referred to as a "lookup table" (LUT) in these situations). When the user
inputs a y, the program searches the LUT for the closest value greater
than y and returns the corresponding x. The code compiles and looks like
it should work, but when you run the program it crashes.

1) Fix the program so that it runs correctly without crashing, of course
keeping the intended precision of the lookup table.

2) Improve the program! The original programmer wasn't very clever. I'm
sure you can find many things that could be done better.
*/

/*
1) The program crashes because of a stack overflow. We are trying to put
a very large array on the stack, but there's not enough space in memory.
We could allocate the array on the heap, but since arrays are evil, a
much better solution is to not use arrays at all. If we simply create
and fill a vector<double> instead, the stack overflow disappears. That's
because the vector stores its elements on the heap. All of the memory
management is done within the vector class, so we don't have to worry
about it.

2) The problem is that we are doing a slow, linear search through the
lookup table every time we call searchInTable. The LUT is ordered, so
we could easily do a binary search instead. That reduces the number of
operations from O(n) to O(log(n)).

You could also switch to a map instead of a vector. That's justifiable
if the time spent building the map is small compared to the total 
running time of the program. Searching in the map is indeed very fast,
but as you will see if you try it, building a map this large takes
about the same time as doing all those linear searches in the vector.

Another approach would be to abandon the lookup table and use other
numerical methods like Newton-Raphson.
*/

const int nPoints = 10000000; //Number of points in the LUT. This sets the LUT precision.
const double xMax = 100;
const int nSearches = 500;

double searchInTable(double y, vector<double>& table) { //Do a binary search!
	vector<double>::iterator it = lower_bound(table.begin(), table.end(), y);
	if(it == table.end()) return 0.; //Value not in table
	return (it - table.begin())*xMax/nPoints; //A difference between two vector iterators returns the number of elements between them
}

int main() {
	vector<double> yVal; //vector to store the y-values
	double x = 0.;
	for(int i = 0; i < nPoints; ++i) {
		yVal.push_back(x*exp(x)); //For each x, calculate y and store it here
		x += xMax/nPoints;
	}
	
	//Now let's search in the LUT many times
	double yMax = xMax*exp(xMax);
	cout << "Inverting y = x*exp(x) " << nSearches << " times..." << endl;
	for(int i = 0; i < nSearches; ++i) {
		searchInTable(rand()*yMax/RAND_MAX, yVal); //RAND_MAX is the maximum integer returned by rand()
		if(i % (nSearches/10) == 0) cout << i << endl; //This is just a progress indicator
	}
	cout << "Done!" << endl;
	
	return 0;
}

