#include <iostream>
#include <vector>
#include <cassert>
using namespace std;

/*
str is a local variable that is deallocated at the end of the function.
str.c_str() returns a pointer to some internal representation of the
string. After the string has been deallocated, the memory that c_str()
points to no longer belongs to the string and can be overwritten by 
other variables.

This means that the program might print "I am a string", and it might
not. It depends on whether or not the memory that used to belong to the
string was overwritten. If the program printed gibberish, try to
decrease the number of int* that you put in vec. Instead of five, fill
the vector with only one int*. Does it print "I am a string" now? That
means the memory that used to belong to the string was overwritten by
the int pointers.
*/
const char* getString() {
	string str("I am a string");
	return str.c_str(); //Return the C-string representation of this string
}

/*
The explanation for this one is the same as above, really. We are 
creating a pointer to the local variable "seven", which we return. But
seven is deallocated at the end of the function and the memory that used
to belong to seven can now be used by other stuff. This means that the
pointer will point to gibberish (or at the very least least, it will
point to something that is not the number 7).
*/
int* getPtrToSeven() {
	int seven = 7;
	int* returnValue = &seven;
	return returnValue; //Return a pointer to the number 7
}

/*
A division between two integers will return a new integer. If the result
of the division it NOT integer, all the decimals are truncated away!
That means 2.5 becomes just 2. Note that the integer isn't rounded, but
truncated. This means, for example, that 5/3 = 1.66 would return 1
rather than 2.
*/

double divideNumbers() {
	int numerator = 5;
	int denominator = 2;
	return numerator/denominator; //Return the result of dividing 5 by 2
}

/*
The difference between a float and a double is that the double uses more
bits to represent a number (64 for a double compared to 32 for a float,
typically). But the fact is that neither a double nor a float can
represent the number 0.123 exactly because it has an infinite decimal
expansion in base 2. The float and the double will approximate 0.123 as
good as they can, and in the end the double gets slightly closer. That's
why they will not compare equal.

This is also why you should be careful when comparing two floating point
numbers to each other. If you know that they have the same precision,
it's usually fine. But if they do not, something like

return (abs(num1 - num2) < 0.00001)

would be more appropriate (0.00001 being grabbed out of thin air, any
sufficiently small number that is still greater than the precision of
a float would do).
*/
bool compareNumbers() {
	float num1 = 0.123;
	double num2 = 0.123;
	return (num1 == num2); //Return true if the numbers are equal, false otherwise
}

/*
Unlike above, these numbers CAN be represented exactly in base 2. I'll
even write it out - It's 1111011. Therefore the double and the float
compare equal and the function returns true;
*/
bool compareOtherNumbers() {
	float num1 = 123;
	double num2 = 123;
	return (num1 == num2); //Return true if the numbers are equal, false otherwise
}

/*
Signed types like int have one bit reserved to represent the sign of the
number (0 for positive, 1 for negative). Unsigned types do not interpret
this bit as a sign, but as "part of the number". Without going into
detail, the very same bits that would represent the number -5 for a 32
bit signed type represent the number 2^32 - 5 = 4294967291 for an
unsigned type. For a 64 bit unsigned type, they represent the number
2^64 - 5 = 18446744073709551611.

size_t is an alias for some unsigned type. It's implementation specific
just what type that is, but you probably got one of the above two
numbers. To get the correct behavior, we could just treat the return
type as an int when we call the function:

int vecSize = sizeMinusTen(vec);
*/
size_t sizeMinusTen(vector<int*> vec) {
	return vec.size() - 10; //Return the size of the vector minus 10
}

int main() {
	const char* testStr = getString(); //Should contain C-string that says "I am a string"
	int* seven = getPtrToSeven(); //Should be a pointer to the number seven
	double twoPointFive = divideNumbers(); //Should be 2.5
	bool isEqual = compareNumbers(); //Should be true
	bool isEqual2 = compareOtherNumbers(); //Should be true
	
	/*
	This one is really tricky. We have created a vector with five copies
	of a pointer to the number 10. That's right - Five copies of the SAME
	pointer! The "new" operator is called only once. It allocates space
	on the heap for an int, and returns a pointer to that number. This
	pointer is then copied five times and placed in the vector.
	An alternative approach, which calls new five times, is shown here.
	*/
	//vector<int*> vec(5, new int(10)); //Make a vector containing five pointers to the number 10
	vector<int*> vec; //Now, make an empty vector and then fill it five times
	for(int i = 0; i < 5; ++i) vec.push_back(new int(10));
	size_t vecSize = sizeMinusTen(vec); //The vector size is 5, so this should be the number -5
	
	cout << "Testing variables" << endl;
	cout << testStr << endl; //Should print "I am a string"
	cout << *seven << endl; //Should print 7
	cout << twoPointFive << endl; //Should print 2.5
	cout << isEqual << endl; //Should print 1
	cout << isEqual2 << endl; //Should print 1
	cout << vecSize << endl; //Should print -5
	
	/*
	In the first iteration of this loop, we correctly print 10. Then we
	call delete and the variable that lived on the heap is deallocated.
	In the second iteration, we try to print and delete a variable that
	has been deallocated. This causes a crash because we can not delete
	the same variable twice!
	*/
	cout << endl << "Testing vector" << endl;
	for(int i = 0; i < vec.size(); ++i) {
		cout << *vec.at(i) << endl; //Should print 10
		delete vec.at(i); //Delete the pointer to avoid memory leak
	}
	
	return 0;
}

